'''
Hierarchy of Things, including Veggies, Kittys, Bunnys, and Sounds.
Each is also a Canvas object in the world module, accessed through
its id attribute.
'''

from utils import *
from memory import *
from sense import *
# from cl import *
# from network import *
from hop import *
from learner import *
from genome import *
from network_view import *

# "Radius" of things
THING_SIDE = 12
# Degrees in an open mouth "arc".
MOUTH_OPEN = 60
# Degrees in a closed mouth "arc".
MOUTH_CLOSED = 18

# Minimum and maximum distances traveled by critters.
SPEEDS = {"Bunny": (15, 30), "Kitty": (15, 30), "Bug": (15, 30), "Anemone": (15, 30)}
# Basic colors for orgs, sounds, and lights.
VEGGIE_RGB = (0, 255, 0)
# VEGGIE_FILL_RGB = (0, 25, 0)
BUNNY_RGB = (200, 150, 80)
KITTY_RGB = (150, 150, 150)
BUG_RGB = (80, 150, 225)
SOUND_RGB = (255, 255, 255)
LIGHT_RGB = (255, 255, 100)

# How many steps a sound survives
SOUND_LIFETIME = 4

# Ways to reproduce
DONT_EVOLVE = 0
COPYING = 1
MATING = 2
LEARN = 3
DUMB_COPYING = 5
NO_REPRO = 4

def strength_color(strength, rgb):
    '''Convert a strength to a color string.'''
    return rgb2color([int(strength * x) for x in rgb])

def get_rect(coords, rad=THING_SIDE):
    '''Get the rectangle around a thing.'''
    return (coords[0] - rad, coords[1] - rad,
            coords[0] + rad, coords[1] + rad)

class Thing(object):
    '''Things in the world with a particular location.'''
    
    # Basic color.
    rgb = (255, 255, 255)
    # Radius of thing.
    s = THING_SIDE
    # Label, used for world objects.
    label = ""
    # Current color.
    color = ""
    # Whether the thing is hiding.
    hidden = False
    # Color of the outline.
    outline_color = "white"
    # Sensory properties.
    sensations = {}

    def __init__(self, world, coords=(0, 0), initial=True,
                 color=True, outline_width=1, child=False):
        '''Find a place for the thing and create its Canvas object.'''
        self.world = world
        self.set_features(child)
        if coords == None and world:
            coords = world.get_vacant_pos()
        # Center of the thing
        self.x = coords[0]
        self.y = coords[1]
        if color:
            self.set_color()
        if world:
            self.make_object(outline_width, self.get_rect())
            self.tag(self.label)
            self.tag("obj")
#            print 'Binding', self.id, 'with describe for', self.label
            world.tag_bind(self.id, "<Button-1>", self.describe)

    def set_features(self, child):
        pass

    def __str__(self):
        '''Print name.'''
        return self.label + str(self.id)

    def make_object(self, outline_width, rect):
        '''Create the Canvas object and assign its id to id.'''
        self.id = self.world.create_oval(rect[0], rect[1], rect[2], rect[3],
                                         width=outline_width, outline=self.outline_color,
                                         fill=self.color)

    def set_color(self):
        '''Set the initial color of the graphical object.'''
        c = rgb2color(self.rgb)
        self.outline_color = c
        self.color = c

    def describe(self, event):
        '''Describe the thing (overridden in subclasses).'''
        print self.__str__()

    def get_rect(self):
        '''Rectangle around the thing.'''
        return (self.x - self.s, self.y - self.s,
                self.x + self.s, self.y + self.s)

    def step(self, verbose=True):
        '''Step the thing once (overridden in subclasses).'''
        pass

    def tag(self, s):
        '''Add s to tags of the Canvas object.'''
        self.world.addtag_withtag(s, self.id)

    def dtag(self, s):
        '''Delete s from the tags of the Canvas object.'''
        self.world.dtag(self.id, s)

    def update_graphics(self):
        pass

class Org(Thing):
    '''Living Things.'''
    
    reproduced = False

    def __init__(self, world, coords, color, genome=None, immortal=True, child=True):
        '''Do Thing init and set age, strength, etc.'''
        Thing.__init__(self, world, coords, color, child=child)
        self.genome = genome
        if child:
            self.age = 0
        else:
            self.age = random.randint(0, self.life_expect / 2)
        self.alive = True
        self.immortal = immortal

    def set_features(self, child):
        '''Assign basic features relevant for reproduction and learning.'''
        self.init_strength(child)
        self.life_expect = self.world.features.get('LIFE_EXPECT', 4000)
        self.max_strength = self.world.features.get('MAX_STRENGTH', 4000)
        self.copy_prob = self.world.features.get('MAX_EXPECTED_CHILDREN', 10.0) / \
                         (self.life_expect * self.max_strength * self.max_strength)
        self.mate_prob = self.world.features.get('MAX_MATE_PROB', .1) / \
                         (self.max_strength**4)
        self.food_value = self.world.features.get('FOOD_VALUE', .012)

    def crowded(self):
        '''Whether the Org overlaps with other Orgs of its type.'''
        ol = self.world.overlapping(thing_id=self.id)
        if ol:
            cls = self.__class__
            for thing in ol:
                if thing.__class__ == cls:
                    return True
        return False

    def copy(self):
        '''Copy this Org, copying the genome if it has one.'''
        return Org(self.world, coords=self.world.get_vacant_pos(), color=self.color, child=True,
                   genome=self.genome.copy() if self.has_genome() else None)

    def has_genome(self):
        '''Does the Org have a genome?'''
        return self.genome is not None

    def step(self, verbose=True):
        '''Step once: age and die or reproduce if appropriate.'''
        self.age += 1
        if self.too_old() and not self.immortal:
            self.world.died_age += 1
            self.tag('kill')
        elif self.strength == 0:
            self.world.starved += 1
            self.tag('kill')
        if not self.world.reproduction == NO_REPRO:
            self.reproduce()

    def reproduce(self):
        '''Decide whether to reproduce and then do it (overridden in subclasses).'''
        if random.random() < COPY_PROB * self.strength * self.strength:
            if self.has_genome():
                self.world.copy.append(self)
            else:
                self.world.born.append(self.offspring_props())
            self.reproduced = True

    def get_older(self, steps):
        """Age the critter by steps time steps."""
        self.age += steps

    def too_old(self):
        '''Die of old age?  Probabilistic function of age.'''
        return random.random() < sigmoid(self.age, self.life_expect, .05)

    def get_sensation(self, kind):
        '''How this org feels to a particular kind of sense.'''
        return self.sensations.get(kind, None)

    def update_graphics(self):
        '''Do non-essential graphics.'''
        self.update_color()

    def init_strength(self, child):
        '''Initialize the strength of the org.'''
#        if child:
#            self.strength = self.world.features.get('CHILD_STRENGTH', 2500)
#        else:
        self.strength = self.world.features.get('INIT_STRENGTH', 2500)

    def strengthen(self, incr):
        """Change the org's strength by incr."""
        tmp = self.strength + incr
        self.strength = min(self.max_strength, max(tmp, 0))

    def set_color(self):
        '''Set the color on the basis of the strength.'''
        strength_prop = min([1.0, self.strength / float(self.max_strength)])
        self.color = strength_color(strength_prop, self.rgb)

    def update_color(self):
        '''Update the color in the Canvas object.'''
        self.set_color()
        if not self.hidden:
            self.world.itemconfigure(self.id, fill=self.color)

    def describe(self, event):
        '''Print out some useful stuff about the org.'''
        print self
        print "age:" + str(self.age),
        print "strength: " + str(self.strength),
        print
        
    def suffer(self, incr):
        '''Get reinforced (overridden in Critter).'''
        self.strengthen(incr)

    def get_chewed(self):
        '''What happens when the Org is chewed (overridden in Critter).'''
        pass

    def get_food_value(self):
        '''How much the Org is worth as food.'''
        return int(self.strength * self.food_value)

    def die(self):
        '''Get rid of the Org, physically and graphically.'''
        del self.world.things[self.id]
        self.world.delete(self.id)
        self.alive = False

    def offspring_props(self):
        '''Properties for offspring of the Org.'''
        return {'kind': getattr(self, '__class__'),
                'name': self.label}

class Critter(Org):
    '''Orgs that can move and take actions.'''
    
    CHEW_RANGE = 8

    food = "dirt"
    mouth_angle = MOUTH_OPEN
    mouth_opened = False
    reinforcement = 0
    
    def __init__(self, world, coords=None, senses=[],
                 learning=False, competitive=False, hopfield=False,
                 # Are senses dynamic
                 dynamic=False,
                 genome=None, immortal=True, child=True,
                 acts_dict={'move': 0, 'turn': 0, 'eat': 0, 'hide': 0},
                 linear=True):
        '''Do Org init and set heading, acts, senses, network, memory, network view; initialize genome.'''
        self.heading = random.randint(0, 360)
        Org.__init__(self, world, coords, color=True, genome=genome, immortal=immortal, child=child)
        # This has to happen before set_senses
        self.dynamic = dynamic
        self.last_state = []
        self.set_reinforcements()
        self.set_acts(acts_dict)
        self.set_senses(senses)
        self.learning = learning
        self.competitive = competitive
        self.hopfield = hopfield
        self.set_network(linear)
        self.set_memory()
        self.set_learner()
        self.mating_refractory = self.world.features.get('MATING_REFRACTORY', 10)
        self.time_since_mating = self.mating_refractory
        if genome is not None:
            self.genome.initialize()
#        if learner is not None:
#            self.learner.initialize()
        self.net_view = None
        self.crowd = False

    def set_reinforcements(self):
        '''Set basic reinforcements necessary to make evolution and reinforcement learning work.'''
        self.step_cost = self.world.features.get('STEP_COST', 0)
        self.move_cost = self.world.features.get('MOVE_COST', -1)
        self.bump_cost = self.world.features.get('BUMP_COST', -10)
        self.turn_cost = self.world.features.get('TURN_COST', -1)
        self.hide_cost = self.world.features.get('HIDE_COST', -1)
        self.say_cost = self.world.features.get('SAY_COST', -1)
        self.eat_cost = self.world.features.get('EAT_COST', -15)
        self.mate_cost = self.world.features.get('MATE_COST', -25)
        self.chewed_cost = self.world.features.get('CHEWED_COST', -20)
        self.crowding_cost = self.world.features.get('CROWDING_COST', -1)

    def copy(self):
        '''Copy the Critter, copying its genome if it has one.'''
        return Critter(self.world, coords=self.world.get_vacant_pos(), color=self.color, child=True,
                       senses=self.senses, genome=self.genome.copy() if self.has_genome() else None,
                       acts_dict={'move': 0, 'turn': 0, 'eat': 0, 'hide': 0})

    def set_senses(self, senses):
        '''Set the senses (from the arg passed in __init__).'''
        self.senses = [sense(self) for sense in senses]
        self.n_senses = len(senses)
        self.sense_length = sum([s.n_features for s in self.senses])
        self.sensible = list(seqs2seq([s.sensations for s in self.senses]))
        if self.dynamic:
            self.last_state = [0 for f in range(self.sense_length)]

    def set_acts(self, acts_dict):
        '''Set the acts (from the dict passed in __init__).'''
        self.act_bias = [(getattr(self, act), weight) for act, weight in acts_dict.items()]
        self.acts = [ab[0] for ab in self.act_bias]
        self.n_acts = len(acts_dict)
        self.action_labels = acts_dict.keys()

    def set_network(self, linear=False):
        '''Set the neural network, using the senses and acts to determine # of inputs and outputs.'''
        self.network = Network(self.__str__(),
                               [Layer('in', self.sense_length, linear=True),
                                Layer('out', self.n_acts, weight_range = 0.0, gain = 4.0,
                                      labels=self.action_labels, linear=linear)],
                               supervised=(self.learning and not self.competitive))

    def set_memory(self):
        '''Set the memory.'''
        self.memory = Memory(self, stochastic=True, learning=self.learning)

    def set_learner(self):
        '''Set the Critter's learner, defaults to None.  Overridden in Bunny.'''
        self.learner = None

    def make_object(self, outline_width, rect):
        '''A critter appears as a filled circle with an empty mouth.'''
        self.id = self.world.create_arc(rect[0], rect[1], rect[2], rect[3],
                                        start=self.heading,
                                        outline=self.outline_color,
                                        extent=360 - MOUTH_CLOSED,
                                        width=1,
                                        fill=self.color)

    def show_network(self):
        """Display the Critter's network."""
        if not self.net_view:
            self.net_view = NNTop(critter=self, network=self.network, title=self.__str__(),
                                  menus=self.competitive)
            self.net_view.paint()

    def describe(self, event):
        '''Print out useful information about the Critter, and show its network.'''
        print self
        print 'heading', self.heading,
        print "age:", self.age,
        print "strength:", self.strength
        if self.has_genome():
            self.genome.show()
        self.show_network()

    def get_internal(self):
        '''A list of internal sensory information.'''
        return [self.strength]
    
    def sensory_state(self):
        '''A vector representing the current input to all senses.'''
        return seqs2seq([s.to_vector() for s in self.senses])

    def run_network(self, sensed, target=[], train=False, lr=None, show=True):
        '''Propagate the sensed vector through the network; train if target.'''
        
        self.network.step((sensed, target) if (self.learning and self.network.supervised) else sensed,
                          train=train, lr=lr)
        if self.net_view and show:
            self.net_view.canvas.step0(target=target)

    def sense_reinforce(self):
        '''The total reinforcement (cost) for all senses.'''
        return sum([s.reinforcement for s in self.senses])

    def step(self, verbose=True):
        """Do various bookkeeping things, decide what to do, and act.  Learn if learning."""
        # Reinitialize reinforcement
        self.reinforcement = 0

        # Graphics
        if self.mouth_opened:
            self.close_mouth()

        # Get sensory input
        state = self.sensory_state()
        # Run the network once, learning if learning is ON and this is a competitive learner
        self.run_network(state, train=(self.competitive and self.learning and self.world.learn))
        # Save the output activations
        outputs = self.network.layers[-1].activations

        # This includes a call to reproduce()
        Org.step(self, verbose=verbose)

        if self.alive:  # the Critter may have died of old age

            # Update reinforcement for sensing cost and cost of step
            self.reinforcement += self.sense_reinforce()
            self.reinforcement += self.step_cost

            # Select an network output index on the basis of the output activations
            action_index = self.memory.decide()

            # Execute decision and increment reinforcement based on what happens
            self.reinforcement += self.memory.output2action(action_index)()

            if self.crowd and self.crowded():
                self.reinforcement += self.crowding_cost

            # Q learning; do this if this is a learner (and not a competitive learner)
            # and learning is ON in the world
            if self.learning and self.world.learn and not self.competitive:
                if self.world.n_steps != 0:
                    self.learner.learn(state, outputs)
                self.learner.last_reinf = self.reinforcement
                self.learner.last_state = state
                self.learner.last_action = action_index

            # Update strength
            self.strengthen(self.reinforcement)

            # Save state as last_state
            if self.dynamic:
                self.last_state = state

            # Graphics
            if verbose:
                self.update_color()

    def die(self):
        """Remove the critter from the world."""
        Org.die(self)
        for s in self.senses:
            s.die()

    def suffer(self, incr):
        '''Get reinforced by incr.'''
        Org.suffer(self, incr)
        self.reinforcement += incr

    def set_mouth_angle(self, ang):
        '''Set the angle for the mouth arc.'''
        self.world.itemconfigure(self.id, extent=360 - ang,
                                  start=self.heading - 360 + ang / 2)

    def set_mouth_heading(self, ang):
        '''Set the heading for the mouth.'''
        self.world.itemconfigure(self.id, start=self.heading - 360 + ang / 2)

    def close_mouth(self):
        """Close the Critter's mouth."""
        self.mouth_opened = False
        self.set_mouth_angle(MOUTH_CLOSED)

    def open_mouth(self):
        """Open the Critter's mouth."""
        self.mouth_opened = True
        self.set_mouth_angle(self.mouth_angle)

    def get_chewed(self):
        '''Get punished for being chewed.'''
        self.suffer(self.chewed_cost)

    ## ACTS

    def say(self):
        '''Make an mmm sound.'''
        self.world.add_sound((self.x, self.y), "mmm")
        return self.say_cost

    def yelp(self):
        '''Make an ouch sound.'''
        self.world.add_sound((self.x, self.y), "ouch")
        return self.say_cost

    def chew(self):
        '''Display chewing.'''
        self.open_mouth()

    def hide(self):
        '''Make the Critter unsensible.'''
        if not self.hidden:
            self.hidden = True
            self.tag("hidden")
            self.world.itemconfigure(self.id, fill="")
        return self.hide_cost

    def unhide(self):
        '''Unhide the Critter.'''
        if self.hidden:
            self.dtag("hidden")
            self.hidden = False

    def eat(self):
        '''Attempt to eat whatever is available.'''
        self.unhide()
        self.chew()
        reinf = self.eat_cost
        eaten = []
        for n in self.get_chewable():
            if isinstance(n, self.food):
                self.world.addtag_withtag('chewed', n.id)
                eaten.append(n)
        if len(eaten) > 0:
            reinf += sum([x.get_food_value() for x in eaten])
        return reinf

    def mouth_end(self):
        '''Coordinates of the point where the mouth opens.'''
        return get_endpoint(self.x, self.y, self.heading, THING_SIDE)

    def get_chewable(self):
        '''Things overlapping with the Critter.'''
        end_x, end_y = self.mouth_end()
        return self.world.overlapping(thing_id=self.id,
                                      rect=(end_x - Critter.CHEW_RANGE, end_y - Critter.CHEW_RANGE ,
                                            end_x + Critter.CHEW_RANGE, end_y + Critter.CHEW_RANGE),
                                      exclude='hidden')

    def turn(self, angle=0):
        '''Turn through the given angle.'''
        self.unhide()
        if angle is 0:
            if random.random() < .5:
                angle = 60
            else:
                angle = 300
        self.heading = (self.heading + angle) % 360
        if self.mouth_opened:
            self.set_mouth_heading(self.mouth_angle)
        else:
            self.set_mouth_heading(MOUTH_CLOSED)
        for s in self.senses:
            s.turn()
        return self.turn_cost

    def turn_move(self):
        '''Turn some random amount and then try to move.'''
        angle = (random.randint(-90, 90) + 360) % 360
        self.heading = (self.heading + angle) % 360
        if self.mouth_opened:
            self.set_mouth_heading(self.mouth_angle)
        else:
            self.set_mouth_heading(MOUTH_CLOSED)
        return self.move()

    def turn_right(self):
        '''Turn 1/6 to the right.'''
        return self.turn(300)

    def turn_left(self):
        '''Turn 1/6 to the left.'''
        return self.turn(60)

    def move(self):
        '''Move forward, distance depending on speed.'''
        self.unhide()
        reinf = self.move_cost
        dists = self.move_dists(self.get_speed())
        old_x = self.x
        old_y = self.y
        # Try to move here
        self.x += dists[0]
        self.y += dists[1]
        # Bump walls?
        if self.bump(old_x, old_y):
            reinf += self.bump_cost
        # Move to the possibly adjusted coords
        self.move_to()
        # Total of cost for attempting move and possibly bumping
        return reinf

    def get_speeds(self):
        '''Min and max speeds for the Critter.'''
        return SPEEDS[self.label]

    def get_speed(self):
        '''Actual speed, based in strength.'''
        speeds = self.get_speeds()
        return self.strength / float(self.max_strength) * (speeds[1] - speeds[0]) + speeds[0]

    def move_dists(self, dist):
        '''x and y distances to move a certain distance.'''
        rad_head = math.radians(360 - self.heading)
        return [dist * math.cos(rad_head),
                dist * math.sin(rad_head)]

    def move_to(self):
        '''Move the Org and its graphical object.'''
        rect = self.get_rect()
        self.world.coords(self.id, rect[0], rect[1], rect[2], rect[3])
        for s in self.senses:
            s.move()

    def bump(self, old_x, old_y):
        """Check whether the move just made bumps into a wall and adjust if so."""
        bumped = False
        # get borders of walls, each a list of 4 coordinates
        for w in self.world.vert_wall_bounds + self.world.horiz_wall_bounds:
            if w[1] < old_y < w[3]:
                # crossed right border from the right
                if self.x <= w[2] < old_x:
                    bumped = True
                    self.x = 2 * w[2] - self.x
                # crossed left border from the left
                elif self.x >= w[0] > old_x:
                    bumped = True
                    self.x = 2 * w[0] - self.x
            if w[0] < old_x < w[2]:
                # crossed bottom border from below
                if self.y <= w[3] < old_y:
                    bumped = True
                    self.y = 2 * w[3] - self.y
                # crossed top border from above
                elif self.y >= w[1] > old_y:
                    bumped = True
                    self.y = 2 * w[1] - self.y
        return bumped
        
class Bunny(Critter):
    '''Herbivores with Touch and Feel sensors.'''
    
    def __init__(self, world, coords=None, child=True):
        self.rgb = BUNNY_RGB
        self.outline_color = rgb2color(self.rgb)
        self.label = "Bunny"
        self.food = Veggie
        self.sensations = {"odor": "gamey", "texture": "fuzzy"}
        learning = world.features['THINGS'][Bunny].get('learn', True)
        evolving = world.features['THINGS'][Bunny].get('evolve', False)
        senses = world.features['THINGS'][Bunny].get('senses', [Touch, Feel])
        Critter.__init__(self, world, coords,
                         acts_dict={'move': 0, 'turn_left': 0, 'turn_right': 0, 'eat': 0, 'hide': 0},
                         senses=senses, linear=True,
                         learning=learning,
                         genome=Genome(self) if evolving else None,
                         immortal=not evolving, child=child)
        self.crowd = True

    def reproduce(self):
        '''Mate or copy with a certain probability, depending on setting in GUI.'''
        if self.world.reproduction == MATING:
            # Reproduce by mating
            if 'mate' in self.world.gettags(self.id):                # Recently mated
                if self.time_since_mating >= self.mating_refractory: # Waited long enough to mate again
                    self.dtag('mate')
                else:
                    self.time_since_mating += 1
            if not 'mate' in self.world.gettags(self.id):            # Overlapping partner?
                overlapping = self.world.overlapping(thing_id=self.id, thing_label=self.label, exclude='mate')
                if overlapping:
                    if len(overlapping) == 1:
                        strongest = overlapping[0]
                    else:
                        strongest = argmax(overlapping, lambda x: getattr(x, 'strength'))

                    if random.random() < self.mate_prob * self.strength**2 * strongest.strength**2:
                        self.tag('mate')
                        strongest.tag('mate')
                        self.time_since_mating = 0
                        strongest.time_since_mating = 0
                        self.world.mates.append((self, strongest))

        elif self.world.reproduction == COPYING and random.random() < self.copy_prob * self.strength * self.strength:
            # Reproduce by copying
            self.world.copy.append(self)
            self.reproduced = True

    def set_learner(self):
        '''Make the Bunny's learner an instance of QLearner.'''
        if self.learning:
#            pass
            self.learner = QLearner(self)
        else:
            Critter.set_learner(self)

    def set_memory(self):
        '''Memory that decides deterministically.'''
        self.memory = Memory(self, stochastic=True, learning=self.learning)

    def copy(self):
        '''Copy the bunny, preserving its genome (with mutation) if evolution is happening.'''
        bunny = Bunny(self.world, coords=self.world.get_vacant_pos(), child=True)
        if self.has_genome():
            if self.world.evolve:
                bunny.genome = self.genome.copy(critter=bunny)
            else:
                bunny.genome = Genome(bunny)
                bunny.genome.initialize()
        return bunny

class Kitty(Critter):
    '''Carnivores with a Smell sensor.'''
    
    def __init__(self, world, coords=None, child=True):
        self.rgb = KITTY_RGB
        self.outline_color = rgb2color(self.rgb)
        self.label = "Kitty"
        self.food = Bunny
        self.sensations = {"odor": "pungent", "texture": "soft"}
        Critter.__init__(self, world, coords, senses=[Smell],
                         acts_dict={'move': 0, 'turn': -0.5, 'eat': 0},
                         child=child)
        ## Rule to hard-wire a smart Kitty
        # Eat something gamey (Bunny)
        self.memory.add_rule(("gamey", (self.eat, 3), (self.turn, -1), (self.move, -1)))

    def reproduce(self):
        """Don't reproduce."""
        pass

    def strengthen(self, incr):
        '''Prevent Kitties from getting stronger or weaker.'''
        pass

class Bug(Critter):
    '''Critters that see Lights and smell Veggies.'''
    
    def __init__(self, world, coords=None, child=True):
        self.rgb = BUG_RGB
        self.outline_color = rgb2color(self.rgb)
        self.label = "Bug"
        self.food = Veggie
        self.sensations = {"odor": "fragrant", "texture": "slippery"}
        self.sigma = world.features.get('KOH_SIGMA', .8)
        self.lr = world.features.get('KOH_LR', .1)
        self.decay = world.features.get('KOH_DECAY', .9995)
        Critter.__init__(self, world, coords, senses=[Sight, Smell],
                         learning=True, competitive=True, dynamic=True,
                         acts_dict={'turn_move': 0},
                         immortal=True, child=child)

    def set_network(self, linear = False):
        '''Set the neural network, using the senses and acts to determine # of inputs and outputs.'''
        self.network = CompetNetwork(self.__str__(),
                                     [Layer('in', self.sense_length),
                                      KohLayer('out', lr=self.lr, sigma=self.sigma, decay=self.decay,
                                               dimensions=(10, 10))])

    def reproduce(self):
        """Don't reproduce."""
        pass

    def strengthen(self, incr):
        '''Prevent Bugs from getting stronger or weaker.'''
        pass

class Veggie(Org):
    '''Plants.'''

    def __init__(self, world, coords=None, initial=True, child=True):
        self.rgb = VEGGIE_RGB
        self.outline_color = rgb2color(self.rgb)
        self.label = "Veggie"
        self.odor = "sweet"
        self.texture = "prickly"
        self.sensations = {"odor": "sweet", "texture": "prickly"}
        Org.__init__(self, world, coords, color=True, immortal=True, child=child)
        self.growth = self.world.features.get('GROWTH', 0)

    def make_object(self, outline_width, rect):
        '''Veggies are empty rectangles.'''
        self.id = self.world.create_rectangle(rect[0], rect[1], rect[2], rect[3],
                                               width=outline_width, outline=self.outline_color,
                                               fill=self.color)

    def init_strength(self, child):
        '''Randomize initial strength.'''
        init = self.world.features.get('INIT_VEGGIE_STRENGTH', 4000)
        self.strength = random.randint(init - 100, init + 100)

    def step(self, verbose=True):
        '''Get older and stronger.'''
        Org.step(self, verbose=verbose)
        if self.alive:   # the Veggie may have died of old age or weakness
            self.strengthen(self.growth)
            if verbose:
                self.update_color()

    def reproduce(self):
        """Don't reproduce."""
        pass

    def get_chewed(self):
        '''Veggies die when they get chewed by Bunnies.'''
        self.die()

class Anemone(Critter):
    '''Herbivores with Smell and Feel sensors.'''

    # Sensory inputs that the anemone should learn along with their tolerances
    special_patterns = (# 0: blue and wavy all around
                        (0, (1,-1,-1,  1,-1,-1,  1,-1,-1,  1,-1,-1,  1,-1,-1,  1,-1,-1,  1,-1)),
                        # 1: green and pointy all around
                        (0, (-1,1,-1,  -1,1,-1,  -1,1,-1,  -1,1,-1,  -1,1,-1,  -1,1,-1,  -1,1)),
                        # 2: orange and wavy all around
                        (0, (-1,-1,1,  -1,-1,1,  -1,-1,1,  -1,-1,1,  -1,-1,1,  -1,-1,1,  1,-1)),
                        # 3: orange and pointy ahead, emptiness behind
                        (1, (-1,-1,1,  -1,-1,1,  -1,-1,1,  -1,-1,-1,  -1,-1,-1,  -1,-1,-1,  -1,1)),
                        # 4: blue and pointy behind, emptiness ahead
                        (1, (-1,-1,-1,  -1,-1,-1,  -1,-1,-1,  1,-1,-1,  1,-1,-1,  1,-1,-1,  -1,1)),
                        # 5: green and wavy ahead, emptiness behind
                        (1, (-1,1,-1,  -1,1,-1,  -1,1,-1,  -1,-1,-1,  -1,-1,-1,  -1,-1,-1,  1,-1)),
                        # 6: green ahead, orange behind, pointy and wavy
                        (2, (-1,1,-1,  -1,1,-1,  -1,1,-1,  -1,-1,1,  -1,-1,1,  -1,-1,1,  1,1)))
    
    def __init__(self, world, coords=None, child=True):
        self.rgb = (255, 255, 0)
        self.outline_color = rgb2color(self.rgb)
        self.label = "Anemone"
        self.food = Plasma
        self.sensations = {"odor": "gamey", "texture": "fuzzy"}
        Critter.__init__(self, world, coords, senses=[Feel, Smell],
                         acts_dict={'turn_move': 0},
                         linear=True, learning=True, competitive=False, hopfield=True, dynamic=False,
                         immortal=True, child=child)

    def reproduce(self):
        pass

    def strengthen(self, incr):
        '''Prevent Bugs from getting stronger or weaker.'''
        pass

    def match_pattern(self, sensed):
        '''Does the sensory state match any of the special patterns?'''
        for index, tol_pat in enumerate(Anemone.special_patterns):
            tolerance, pattern = tol_pat
            if hamdist(pattern, sensed) <= tolerance:
                print 'Matched pattern', index
                return index
        return False

    def step(self, verbose=True):
        """Override Critter.step because a lot is simplified."""

        # Get sensory input, replacing 0s with -1s
        state = [(1 if x == 1 else -1) for x in self.sensory_state()]
        if self.world.learn and self.match_pattern(state) is not False:
            # Do Hopfield learning on the state pattern
            self.run_network(state, train=True)
        else:
            # Just run the network
            self.run_network(state, train=False)

        # This includes a call to reproduce()
        Org.step(self, verbose=verbose)

        # Only one action possible; don't bother to decide
        self.acts[0]()

    def set_network(self, linear=False):
        '''Set the neural network, using the senses to determine # of units.'''
        self.network = HopfieldNetwork(self.__str__(), self.sense_length)
    
class Plasma(Org):
    '''A large blob with a specified odor (shape) and texture (color).'''

    radius = 60
    textures = (("prickly", (100, 50, 250)), ("slippery", (0, 150, 0)), ("slimy", (255, 100, 0)))
    odors = (("pungent", 1.2), ("fragrant", 1.7))

    def __init__(self, world, coords=None, initial=True, odor="sweet", child=False):
        self.label = "Plasma"
        self.sensations = {}
        self.sensations['odor'] = odor
        self.set_texture_odor()
        Org.__init__(self, world, coords, color=True, immortal=True, child=True)

    def set_texture_odor(self):
        '''Set random texture and odor, controlling color and shape.'''
        texture, self.rgb = random.choice(Plasma.textures)
        odor, self.rad_div = random.choice(Plasma.odors)
        self.sensations['odor'] = odor
        self.sensations['texture'] = texture
        self.outline_color = rgb2color(self.rgb)

    def get_rect(self):
        '''Rectangle around the thing. Overrides Thing.get_rect because of greater radius.'''
        return (self.x - Plasma.radius, self.y - Plasma.radius,
                self.x + Plasma.radius, self.y + Plasma.radius)

    def make_object(self, outline_width, rect):
        '''Create the shapes depending on the small (self.rad_div).'''
        self.id = self.world.create_polygon(# NW
                                            self.x - Plasma.radius, self.y - Plasma.radius,
                                            self.x - Plasma.radius/2, self.y - Plasma.radius/self.rad_div,
                                            # N
                                            self.x, self.y - Plasma.radius,
                                            self.x + Plasma.radius/2, self.y - Plasma.radius/self.rad_div,
                                            # NE
                                            self.x + Plasma.radius, self.y - Plasma.radius,
                                            self.x + Plasma.radius/self.rad_div, self.y - Plasma.radius/2,
                                            # E
                                            self.x + Plasma.radius, self.y,
                                            self.x + Plasma.radius/self.rad_div, self.y + Plasma.radius/2,
                                            # SE
                                            self.x + Plasma.radius, self.y + Plasma.radius,
                                            self.x + Plasma.radius/2, self.y + Plasma.radius/self.rad_div,
                                            # S
                                            self.x, self.y + Plasma.radius,
                                            self.x - Plasma.radius/2, self.y + Plasma.radius/self.rad_div,
                                            # SW
                                            self.x - Plasma.radius, self.y + Plasma.radius,
                                            self.x - Plasma.radius/self.rad_div, self.y + Plasma.radius/2,
                                            # W
                                            self.x - Plasma.radius, self.y,
                                            self.x - Plasma.radius/self.rad_div, self.y - Plasma.radius/2,
                                            width=outline_width, outline='',
                                            smooth=1, fill=self.color)

    def step(self, verbose=True):
        """Don't do anything."""
        pass

    def reproduce(self):
        """Don't reproduce."""
        pass

    def get_chewed(self):
        pass

class Light(Thing):
    '''Lights are permanent things with fixed positions, colors, and brightnesses.'''

    radius = 4

    def __init__(self, world, coords, color, brightness):
        self.sensations = {"light": "bright"}
        self.fill = color
        self.brightness = brightness
        Thing.__init__(self, world, coords)

    def make_object(self, outline_width, rect):
        '''Create the Canvas object and assign its id to id.'''
        self.id = self.world.create_oval(rect[0], rect[1], rect[2], rect[3],
                                          width=0, outline='',
                                          fill=self.fill)
    def get_rect(self):
        '''Rectangle around the light.'''
        rad = self.brightness * Light.radius
        return (self.x - rad, self.y - rad,
                self.x + rad, self.y + rad)

class Sound(Thing):
    '''Sounds are temporary Things.'''

    age = 0
    kill = False

    def __init__(self, world, coords, text):
        self.rgb = SOUND_RGB
        self.outline_color = rgb2color(self.rgb)
        self.label = "sound"
        self.text = text
        self.lifetime = SOUND_LIFETIME
        self.sensations = {"sound": text}
        Thing.__init__(self, world, coords)

    def make_object(self, outline_width, rect):
        '''Create the graphical representation: a text object.'''
        self.id = self.world.create_text(self.x, self.y,
                                          text=self.text,
                                          font=("Courier", "10"),
                                          fill=self.color)

    def step(self, verbose=True):
        """Age the sound and get rid of it if it's too old."""
        self.age += 1
        if self.age > self.lifetime:
            self.die()

    def die(self):
        '''Get rid of the graphical object and the thing itself.'''
        self.kill = True
        self.world.delete(self.id)
